from __future__ import unicode_literals
from django.db import models, connection
from .private_media import *
from django.contrib.auth.models import User
from django.core.validators import MaxValueValidator, MinValueValidator
import uuid

# ENUM choices
DISCOUNT_CHOICE = (
    ('Percentage', 'Percentage'),
    ('Amount', 'Amount'),
)

KANBAN_BOARD_STATUS_CHOICE = (
    ('Open','Open'),
    ('Closed','Closed'),
)

PAGE_LAYOUT = (
    ('Landscape','Landscape'),
    ('Portrait','Portrait'),
)

PERMISSION_LEVEL = (
    (0, 'No Permission'),
    (1, 'Read Only'),
    (2, 'Edit Only'),
    (3, 'Add and Edit'),
    (4, 'Full Permission'),
)

PERMISSION_BOOLEAN = (
    (0, 'No Permission'),
    (1, 'Has Permission'),
)

PROJECT_STATUS_CHOICE = (
    ('Backlog', 'Backlog'),
    ('Blocked', 'Blocked'),
    ('In Progress', 'In Progress'),
    ('Test/Review', 'Test/Review'),
    ('Closed', 'Closed'),
)

QUOTE_APPROVAL_STATUS = (
    ('REJECTED', 'REJECTED'),
    ('DRAFT', 'DRAFT'),
    ('APPROVED', 'APPROVED'),
)

RATING_SCORE = (
    (1, '1 Star'),
    (2, '2 Star'),
    (3, '3 Star'),
    (4, '4 Star'),
    (5, '5 Star'),
)

RFC_APPROVAL = (
    (1,'Waiting'),
    (2,'Approved'),
    (3,'Rejected'),
    (4,'Cancel'),
)

RFC_IMPACT = (
    (3,'High'),
    (2,'Medium'),
    (1,'Low'),
)

RFC_PRIORITY = (
    (4,'Critical'),
    (3,'High'),
    (2,'Medium'),
    (1,'Low'),
)

RFC_RISK = (
    (5,'Very High'),
    (4,'High'),
    (3,'Moderate'),
    (2,'Low'),
    (1,'None'),
)

RFC_STATUS = (
    (1,'Draft'),
    (2,'Waiting for approval'),
    (3,'Approved'),
    (4,'Started'),
    (5,'Finished'),
    (6,'Rejected'),
)

RFC_TYPE = (
    (4,'Emergency'),
    (3,'High'),
    (2,'Medium'),
    (1,'Low'),
)


WANT_CHOICE=(
    ('0','Do not want to do'),
    ('1','Want to do'),
)
SKILL_CHOICE=(
    ('0','Can not do'),
    ('1','Willing to learn'),
    ('2','Knows a little'),
    ('3','Knows a lot'),
    ('4','Proficient'),
)

WEBSITE_SOURCE=(
    ('Twitter','Twitter'),
    ('Facebook','Facebook'),
    ('Github','Github'),
    ('Gitlab','Gitlab'),
    ('Website','Website'),
    ('LinkedIn','LinkedIn'),
    ('Staff Page','Staff page'),
    ('Other','Other'),
)

# List of tables - in alphabetical order
class about_user(models.Model):
    about_user_id=models.AutoField(primary_key=True)
    about_user_text=models.TextField()
    user=models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "about_user"



"""
Contact History is a simple form that user will fill out every time they
have some form of contact with the customer. This table will store both
contact history for customer and Organisations. The customer field in
this instance is not required, and implies that the contact history is 
applied to the organisation. The organisation field will fill out automatically
when a user applies it to a customer. :)
"""


class contact_history(models.Model):
    contact_history_id = models.AutoField(primary_key=True)
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    contact_type = models.ForeignKey(
        'list_of_contact_type',
        on_delete=models.CASCADE,
    )
    contact_date = models.DateTimeField()
    contact_history = models.TextField('contact_history')
    document_key = models.ForeignKey(
        'document',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "contact_history"


class bug(models.Model):
    bug_id = models.AutoField(primary_key=True)
    bug_client = models.ForeignKey(
        'bug_client',
        on_delete=models.CASCADE,
    )
    bug_code = models.CharField(max_length=255)  # Just stores the code of the bug
    bug_description = models.TextField()
    bug_status = models.CharField(max_length=50)  # Updated manually?
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    requirement = models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey \
        (User,
         on_delete=models.CASCADE,
         related_name='%(class)s_change_user',
         )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.bug_description)

    class Meta:
        db_table = "bug"


class bug_client(models.Model):
    bug_client_id = models.AutoField(primary_key=True)
    bug_client_name = models.CharField(max_length=50)
    list_of_bug_client = models.ForeignKey(
        'list_of_bug_client',
        on_delete=models.CASCADE,
    )
    bug_client_url = models.URLField()
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey \
        (User,
         on_delete=models.CASCADE,
         related_name='%(class)s_change_user',
         )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.bug_client_name)

    class Meta:
        db_table = "bug_client"


class campus(models.Model):
    campus_id = models.AutoField(primary_key=True)
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    campus_nickname = models.CharField(max_length=100)
    campus_phone = models.CharField(
        max_length=20,
        null=True
    )
    campus_fax = models.CharField(
        max_length=20,
        null=True
    )
    campus_address1 = models.CharField(
        max_length=255,
        null=True
    )
    campus_address2 = models.CharField(
        max_length=255,
        null=True
    )
    campus_address3 = models.CharField(
        max_length=255,
        null=True
    )
    campus_suburb = models.CharField(max_length=50)
    campus_region = models.ForeignKey(
        'list_of_country_region',
        on_delete=models.CASCADE,
    )
    campus_postcode = models.CharField(
        max_length=10,
        null=True,
        blank=True,
    )
    campus_country = models.ForeignKey(
        'list_of_country',
        on_delete=models.CASCADE,
    )
    campus_longitude = models.DecimalField(
        decimal_places=13,
        max_digits=16,
        null=True,  # If use has no mapping software, we want to leave this blank
        blank=True,
    )
    campus_latitude = models.DecimalField(
        decimal_places=13,
        max_digits=16,
        null=True,  # If use has no mapping software, we want to leave this blank
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.campus_nickname)

    class Meta:
        db_table = "campus"



class change_task(models.Model):
    change_task_id=models.AutoField(primary_key=True)
    request_for_change=models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
    )
    change_task_title=models.CharField(
        max_length=255,
    )
    change_task_description=models.TextField(
        blank=True,
        null=True,
    )
    change_task_start_date=models.DateTimeField()
    change_task_end_date=models.DateTimeField()
    change_task_seconds=models.BigIntegerField(
        default=0,
    )
    change_task_assigned_user=models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='change_assigned_user',
    )
    change_task_qa_user=models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='change_qa_user',
    )
    change_task_required_by=models.CharField(
        max_length=255,
        default='Stakeholder(s)',
    )
    change_task_status=models.IntegerField(
        choices=RFC_STATUS, #Similar FLOW to RFC
    )
    is_downtime=models.BooleanField(
        default=False,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str('$' + str(self.change_task_title))

    class Meta:
        db_table = "change_task"


class change_task_block(models.Model):
    change_task_blocks = models.ForeignKey(
        change_task,
        on_delete=models.CASCADE,
        related_name='change_task_blocks',
    )
    blocked_change_task = models.ForeignKey(
        change_task,
        on_delete=models.CASCADE,
        related_name='blocked_change_task',
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str('$' + str(self.change_task_title))

    class Meta:
        db_table = "change_task_block"


class cost(models.Model):
    cost_id = models.AutoField(primary_key=True)
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    cost_description = models.CharField(max_length=255, )
    cost_amount = models.DecimalField(
        max_digits=19,
        decimal_places=2
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str('$' + str(self.cost_amount))

    class Meta:
        db_table = "cost"


class customer(models.Model):
    customer_id = models.AutoField(primary_key=True)
    customer_title = models.ForeignKey(
        'list_of_title',
        on_delete=models.CASCADE,
    )
    customer_first_name = models.CharField(max_length=50)
    customer_last_name = models.CharField(max_length=50)
    customer_email = models.CharField(max_length=200)
    customer_profile_picture = models.ImageField(
        blank=True,
        null=True,
        upload_to='profile_pictures'
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(
            str(self.customer_id)
            + ' - '
            + self.customer_first_name
            + ' '
            + self.customer_last_name
        )

    class Meta:
        db_table = "customer"


class customer_campus(models.Model):
    customer_campus_id = models.AutoField(primary_key=True)
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
    )
    campus = models.ForeignKey(
        'campus',
        on_delete=models.CASCADE,
    )
    customer_phone = models.CharField(max_length=20)
    customer_fax = models.CharField(max_length=20)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "customer_campus"


class document(models.Model):
    document_key = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        primary_key=True,
    )
    document_description = models.CharField(max_length=255)
    document_url_location = models.TextField(
        # Contains URLS
        null=True,
        blank=True,
    )
    document = models.FileField(
        blank=True,
        null=True,
        storage=File_Storage(),
    )
    whiteboard = models.ForeignKey(
        'whiteboard',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "document"

    def __str__(self):
        return str(self.document_description)


class document_permission(models.Model):
    document_permisssion_id = models.AutoField(primary_key=True)
    document_key = models.ForeignKey(
        'document',
        on_delete=models.CASCADE,
    )
    project = models.ForeignKey(
        'project',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    task = models.ForeignKey(
        'task',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    organisation = models.ForeignKey(
        'organisation',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    customer = models.ForeignKey(
        'customer',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    opportunity = models.ForeignKey(
        'opportunity',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    whiteboard = models.ForeignKey(
        'whiteboard',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    requirement = models.ForeignKey(
        'requirement',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    requirement_item=models.ForeignKey(
        'requirement_item',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    request_for_change = models.ForeignKey(
        'request_for_change',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )
    folder = models.ForeignKey(
        'folder',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "document_permission"


class email_contact(models.Model):
    email_contact_id = models.AutoField(primary_key=True)
    email_content = models.ForeignKey(
        'email_content',
        on_delete=models.CASCADE,
    )
    to_customer = models.ForeignKey(
        'customer',
        related_name='%(class)s_to_customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    cc_customer = models.ForeignKey(
        'customer',
        related_name='%(class)s_cc_customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    bcc_customer = models.ForeignKey(
        'customer',
        related_name='%(class)s_bcc_customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    opportunity = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    quotes = models.ForeignKey(
        'quote',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    is_private = models.BooleanField(default=False)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "email_contact"


class email_content(models.Model):
    email_content_id = models.AutoField(primary_key=True)
    email_subject = models.CharField(max_length=255)
    email_content = models.TextField('email_content')
    is_private = models.BooleanField(default=False)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "email_content"


class folder(models.Model):
    folder_id = models.AutoField(primary_key=True)
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    customer=models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation=models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement=models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement_item = models.ForeignKey(
        'requirement_item',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    request_for_change = models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    folder_description = models.CharField(max_length=255)
    parent_folder = models.ForeignKey(
        'self',
        blank=True,
        null=True,
        on_delete=models.CASCADE
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.folder_description)

    class Meta:
        db_table = "folder"


class group(models.Model):
    group_id = models.AutoField(primary_key=True)
    group_name = models.CharField(
        max_length=50,
    )
    parent_group = models.ForeignKey(
        "self",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def natural_key(self):
        return (
            self.group_id,
            self.group_name
        )

    def __str__(self):
        return str(self.group_name)

    class Meta:
        db_table = "group"


class group_manager(models.Manager):
    def get_by_natural_key(
            self,
            group_id,
            group_name
    ):
        return self.get(
            group_id=group_id,
            group_name=group_name
        )


class group_permission(models.Model):
    group_permission_id = models.AutoField(primary_key=True)
    permission_set = models.ForeignKey(
        'permission_set',
        on_delete=models.CASCADE,
    )
    group = models.ForeignKey(
        'group',
        on_delete=models.CASCADE
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.permission_set)

    class Meta:
        db_table = "group_permission"


class kanban_board(models.Model):
    kanban_board_id = models.AutoField(primary_key=True)
    kanban_board_name = models.CharField(max_length=255)
    requirement = models.ForeignKey(
        'requirement',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    kanban_board_status= models.CharField(
        max_length=10,
        choices=KANBAN_BOARD_STATUS_CHOICE,
        default="Open",
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "kanban_board"

    def __str__(self):
        return str(self.kanban_board_name)


class kanban_card(models.Model):
    kanban_card_id = models.AutoField(primary_key=True)
    kanban_card_text = models.CharField(max_length=255)
    kanban_card_description = models.TextField(
        blank=True,
        null=True,
    )
    kanban_card_sort_number = models.IntegerField()
    kanban_level = models.ForeignKey(
        'kanban_level',
        on_delete=models.CASCADE,
    )
    kanban_column = models.ForeignKey(
        'kanban_column',
        on_delete=models.CASCADE,
    )
    kanban_board = models.ForeignKey(
        'kanban_board',
        on_delete=models.CASCADE,
    )
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    requirement = models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "kanban_card"

    def __str__(self):
        return str(self.kanban_card_text)


class kanban_column(models.Model):
    kanban_column_id = models.AutoField(primary_key=True)
    kanban_column_name = models.CharField(max_length=255)
    kanban_column_sort_number = models.IntegerField()
    kanban_board = models.ForeignKey(
        'kanban_board',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "kanban_column"

    def __str__(self):
        return str(self.kanban_column_name)


# class kanban_comment(models.Model):
#     kanban_comment_id = models.AutoField(primary_key=True)
#     kanban_comment = models.TextField()
#     kanban_board = models.ForeignKey(
#         'kanban_board',
#         on_delete=models.CASCADE,
#         null=True,
#         blank=True,
#     )
#     kanban_card = models.ForeignKey(
#         'kanban_card',
#         on_delete=models.CASCADE,
#         null=True,
#         blank=True,
#     )
#     user = models.ForeignKey(
#         User,
#         on_delete=models.CASCADE,
#         null=True
#     )
#     user_infomation = models.CharField(max_length=255)
#     date_created = models.DateTimeField(auto_now_add=True)
#     date_modified = models.DateTimeField(auto_now=True)
#     change_user = models.ForeignKey(
#         User,
#         on_delete=models.CASCADE,
#         related_name='%(class)s_change_user'
#     )
#     is_deleted = models.BooleanField(
#         default=False,
#     )
#
#     class Meta:
#         db_table = "kanban_comment"
#
#     def __str__(self):
#         return str(self.kanban_comment)




class kanban_level(models.Model):
    kanban_level_id = models.AutoField(primary_key=True)
    kanban_level_name = models.CharField(max_length=255)
    kanban_level_sort_number = models.IntegerField()
    kanban_board = models.ForeignKey(
        'kanban_board',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "kanban_level"

    def __str__(self):
        return str(self.kanban_level_name)


class kudos(models.Model):
    kudos_key = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        primary_key=True,
    )
    kudos_rating = models.IntegerField(
        choices=RATING_SCORE,
        default=0,
    )
    improvement_note=models.TextField(
        blank=True,
        null=True,
    )
    liked_note=models.TextField(
        blank=True,
        null=True,
    )
    extra_kudos=models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    submitted_kudos=models.BooleanField(
        default=False,
    )
    project=models.ForeignKey(
        'project',
        on_delete=models.CASCADE,

    )
    customer=models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "kudos"


class list_of_amount_type(models.Model):
    amount_type_id = models.AutoField(primary_key=True)
    amount_type_description = models.CharField(max_length=20)
    list_order = models.IntegerField(unique=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.amount_type_description)

    class Meta:
        db_table = "list_of_amount_type"
        ordering = ['list_order']


class list_of_bug_client(models.Model):
    list_of_bug_client_id = models.AutoField(primary_key=True)
    bug_client_name = models.CharField(max_length=50)
    bug_client_api_url = models.CharField(max_length=255)

    # The different API commands
    api_open_bugs = models.CharField(max_length=255)  # Find all open bugs
    api_search_bugs = models.CharField(max_length=255)  # Search command
    api_bug = models.CharField(max_length=255)  # Get that particular bug information - direct link to bug

    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.bug_client_name)

    class Meta:
        db_table = "list_of_bug_client"


class list_of_currency(models.Model):
    currency_id = models.AutoField(primary_key=True)
    currency_description = models.CharField(max_length=20)
    currency_short_description = models.CharField(max_length=4)
    list_order = models.IntegerField(unique=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.currency_description)

    class Meta:
        db_table = "list_of_currency"


class list_of_contact_type(models.Model):
    contact_type_id = models.AutoField(primary_key=True)
    contact_type = models.CharField(max_length=50)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.contact_type)

    class Meta:
        db_table = "list_of_contact_type"


class list_of_country(models.Model):
    country_id = models.CharField(primary_key=True, max_length=2)
    country_name = models.CharField(max_length=50)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.country_name)

    class Meta:
        db_table = "list_of_country"


class list_of_country_region(models.Model):
    region_id = models.AutoField(primary_key=True)
    country = models.ForeignKey(
        'list_of_country',
        on_delete=models.CASCADE,
    )
    region_name = models.CharField(max_length=150)
    region_type = models.CharField(
        max_length=80,
        null=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='%(class)s_change_user', blank=True,
                                    null=True)
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.region_name)

    class Meta:
        db_table = "list_of_country_region"


class list_of_lead_source(models.Model):
    lead_source_id = models.AutoField(primary_key=True)
    lead_source_description = models.CharField(max_length=20)
    list_order = models.IntegerField(unique=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.lead_source_description)

    class Meta:
        db_table = "list_of_lead_source"


class list_of_opportunity_stage(models.Model):
    opportunity_stage_id = models.AutoField(primary_key=True)
    opportunity_stage_description = models.CharField(max_length=50)
    probability_success = models.DecimalField(
        max_digits=3,
        decimal_places=0,
    )
    list_order = models.IntegerField(unique=True)
    opportunity_closed = models.BooleanField(
        default=False,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.opportunity_stage_description)

    class Meta:
        db_table = "list_of_opportunity_stage"
        ordering = ['list_order']


class list_of_quote_stage(models.Model):
    quote_stage_id = models.AutoField(primary_key=True)
    quote_stage = models.CharField(
        max_length=50,
        unique=True,
    )

    is_invoice = models.BooleanField(
        default=False,
    )
    quote_closed = models.BooleanField(
        default=False,
    )
    sort_order = models.IntegerField(unique=True, auto_created=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.quote_stage)

    class Meta:
        db_table = "list_of_quote_stage"


class list_of_requirement_item_status(models.Model):
    requirement_item_status_id = models.AutoField(primary_key=True)
    requirement_item_status = models.CharField(
        max_length=100,
    )
    status_is_closed = models.BooleanField(
        default=False,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True,
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_item_status)

    class Meta:
        db_table = "list_of_requirement_item_status"


class list_of_requirement_item_type(models.Model):
    requirement_item_type_id = models.AutoField(primary_key=True)
    requirement_item_type = models.CharField(
        max_length=100,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True,
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_item_type)

    class Meta:
        db_table = "list_of_requirement_item_type"


class list_of_requirement_status(models.Model):
    requirement_status_id = models.AutoField(primary_key=True)
    requirement_status = models.CharField(
        max_length=50,
    )
    requirement_status_is_closed = models.BooleanField(
        default=False,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True,
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_status)

    class Meta:
        db_table = "list_of_requirement_status"


class list_of_requirement_type(models.Model):
    requirement_type_id = models.AutoField(primary_key=True)
    requirement_type = models.CharField(
        max_length=100,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True,
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_type)

    class Meta:
        db_table = "list_of_requirement_type"


class list_of_tax(models.Model):
    tax_id = models.AutoField(primary_key=True)
    tax_amount = models.DecimalField(
        max_digits=6,
        decimal_places=4,
    )
    tax_description = models.CharField(
        max_length=50,
        blank=True,
        null=True,
    )  # Incase the customer wants to place a name against the tax
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True,
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.tax_amount)  # No need to encode as it is a decimal point

    class Meta:
        db_table = "list_of_tax"


class list_of_title(models.Model):
    title_id = models.AutoField(primary_key=True)
    title = models.CharField(max_length=10)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.title)

    class Meta:
        db_table = "list_of_title"


class nearbeach_option(models.Model):
    """
    This table will store the options for NearBeach. These options will have a new row each time a new option is created
    There does not need to be a is_deleted function
    """
    nearbeach_option_id=models.AutoField(primary_key=True)
    story_point_hour_min=models.IntegerField(
        default=4,
    )
    story_point_hour_max=models.IntegerField(
        default=10,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user',
        blank=True,
        null=True
    )

    def __str__(self):
        return str(self.nearbeach_option_id)

    class Meta:
        db_table = "nearbeach_option"


class object_assignment(models.Model):
    """
    Object permissions is the centralised permissions for all objects
    - Opportunity
    - Quote
    - Requirement
    - Project
    - Task
    - Kanban board
    - Request for change

    These permission are only "ACCESS" permissions. The user/group's over riding permissions determine if the user
    can add, edit etc.
    """
    object_assignment_id=models.AutoField(primary_key=True)
    assigned_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_assigned_user',
        null=True,
        blank=True,
    )
    group_id=models.ForeignKey(
        'group',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    opportunity=models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    quote = models.ForeignKey(
        'quote',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement = models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement_item = models.ForeignKey(
        'requirement_item',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    kanban_board = models.ForeignKey(
        'kanban_board',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    request_for_change = models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    whiteboard = models.ForeignKey(
        'whiteboard',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "object_assignment"

class object_note(models.Model):
    object_note_id = models.AutoField(primary_key=True)
    object_note = models.TextField(
        blank=False,
        null=False,
    )
    kanban_card = models.ForeignKey(
        'kanban_card',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    opportunity = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    quote = models.ForeignKey(
        'quote',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement = models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    requirement_item = models.ForeignKey(
        'requirement_item',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    request_for_change = models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "object_note"


class opportunity(models.Model):
    opportunity_id = models.AutoField(primary_key=True)
    opportunity_name = models.CharField(max_length=255)
    opportunity_description = models.TextField('oppertunity_description')
    currency = models.ForeignKey(
        'list_of_currency',
        on_delete=models.CASCADE,
    )
    opportunity_amount = models.DecimalField(
        max_digits=12,
        decimal_places=2
    )  # Turn into a number widget
    amount_type = models.ForeignKey(
        'list_of_amount_type',
        on_delete=models.CASCADE
    )
    opportunity_expected_close_date = models.DateTimeField()
    opportunity_stage = models.ForeignKey(
        'list_of_opportunity_stage',
        on_delete=models.CASCADE
    )
    opportunity_success_probability = models.IntegerField()  # Between 0% and 100%
    lead_source = models.ForeignKey(
        'list_of_lead_source',
        on_delete=models.CASCADE
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    user_id = models.ForeignKey(User, on_delete=models.CASCADE)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "opportunities"

"""
class opportunity_connection(models.Model):
    opportunity_connection_id = models.AutoField(primary_key=True)
    opportunity=models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
    )
    customer=models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation=models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "organisation_connection"
        """

class organisation(models.Model):
    organisation_id = models.AutoField(primary_key=True)
    organisation_name = models.CharField(max_length=255)
    organisation_website = models.CharField(max_length=50)
    organisation_email = models.CharField(max_length=100)
    organisation_profile_picture = models.ImageField(
        blank=True,
        null=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.organisation_name)

    class Meta:
        db_table = "organisation"



class permission_set_manager(models.Manager):
    def get_by_natural_key(
            self,
            permission_set_id,
            permission_set_name,
            administration_assign_user_to_group,
            administration_create_group,
            administration_create_permission_set,
            administration_create_user,
            bug_client,
            customer,
            email,
            kanban,
            kanban_card,
            organisation,
            project,
            request_for_change,
            requirement,
            task,
            document,
            kanban_comment,
            project_history,
            task_history,
    ):
        return self.get(
            permission_set_id=permission_set_id,
            permission_set_name=permission_set_name,
            administration_assign_user_to_group=administration_assign_user_to_group,
            administration_create_group=administration_create_group,
            administration_create_permission_set=administration_create_permission_set,
            administration_create_user=administration_create_user,
            bug=bug,
            bug_client=bug_client,
            customer=customer,
            email=email,
            kanban=kanban,
            kanban_card=kanban_card,
            opportunity=opportunity,
            organisation=organisation,
            project=project,
            quote=quote,
            request_for_change=request_for_change,
            requirement=requirement,
            task=task,
            document=document,
            kanban_comment=kanban_comment,
            project_history=project_history,
            task_history=task_history,
        )


class permission_set(models.Model):
    objects = permission_set_manager()

    permission_set_id = models.AutoField(primary_key=True)
    permission_set_name = models.CharField(
        max_length=255,
        #unique=True, #issue when we delete previous permission sets
    )
    # ADMINISTRATION PERMISSIONS
    administration_assign_user_to_group = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    administration_create_group = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    administration_create_permission_set = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    administration_create_user = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    bug_client = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    customer = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    kanban = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    kanban_card = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    organisation = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    project = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    request_for_change = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    requirement = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0
    )
    task = models.IntegerField(
        choices=PERMISSION_LEVEL,
        default=0,
    )
    """
    ADDITIVE permission
    ~~~~~~~~~~~~~~~~~~~~
    Designed to add on extra abilities to those user who have "READ ONLY" for certain modules.
    If a user has the ability to "EDIT" for any of these modules, then this section does not 
    need to be populated with data.
    """
    document = models.IntegerField(
        choices=PERMISSION_BOOLEAN,
        default=0,
    )
    kanban_comment = models.IntegerField(
        choices=PERMISSION_BOOLEAN,
        default=0,
    )
    project_history = models.IntegerField(
        choices=PERMISSION_BOOLEAN,
        default=0,
    )
    task_history = models.IntegerField(
        choices=PERMISSION_BOOLEAN,
        default=0,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def natural_key(self):
        return (
            self.permission_set_id,  # 0
            self.permission_set_name,  # 1
            self.administration_assign_user_to_group,  # 2
            self.administration_create_group,  # 3
            self.administration_create_permission_set,  # 4
            self.administration_create_user,  # 5
            self.assign_campus_to_customer,  # 6
            self.associate_project_and_task,  # 7
            self.customer,  # 8
            self.opportunity,  # 11
            self.organisation,  # 12
            self.organisation_campus,  # 13
            self.project,  # 14
            self.requirement,  # 15
            self.requirement_link,  # 16
            self.task,  # 17
            self.document,  # 18
            self.contact_history,  # 19
            self.project_history,  # 20
            self.task_history,  # 21
        )

    def __str__(self):
        return str(self.permission_set_name)

    class Meta:
        db_table = "permission_set"


class project(models.Model):
    project_id = models.AutoField(primary_key=True)
    project_name = models.CharField(max_length=255)
    project_description = models.TextField('project_description')
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    #Only fill this field out if there are no organisation
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    project_start_date = models.DateTimeField()
    project_end_date = models.DateTimeField()
    project_status = models.CharField(
        max_length=15,
        choices=PROJECT_STATUS_CHOICE,
        default='New'
    )
    project_story_point_min = models.IntegerField(default=1)
    project_story_point_max = models.IntegerField(default=4)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.project_name)

    class Meta:
        db_table = "project"


class project_customer(models.Model):
    project_customer_id = models.AutoField(primary_key=True)
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
    )
    customer_description = models.CharField(
        max_length=255,
        null=True,
        blank=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "project_customer"


"""
class project_group(models.Model):
    project_id = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
    )
    group_id = models.ForeignKey(
        'group',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "project_group"
"""

"""
class project_history(models.Model):
    project_history_id = models.AutoField(primary_key=True)
    project_id = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
    )
    user_id = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        null=True
    )
    user_infomation = models.CharField(max_length=255)
    project_history = models.TextField('project_history')
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.project_id)

    class Meta:
        db_table = "project_history"
"""

"""
class project_opportunity(models.Model):
    project_opprtunity_id = models.AutoField(primary_key=True)
    project_id = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
    )
    opportunity_id = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
    )
    opportunity_description = models.CharField(
        max_length=255,
        null=True,
        blank=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "project_opportunity"
"""

"""
class project_stage(models.Model):
    project_stage_id = models.AutoField(primary_key=True)
    project_id = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
    )
    stage_id = models.ForeignKey(
        'stage',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "project_stage"
"""

"""
class project_task(models.Model):
    project_task_id = models.AutoField(primary_key=True)
    project_id = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        db_column='project_id'
    )
    task_id = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        db_column='task_id'
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "project_task"
"""

class quote(models.Model):
    quote_id = models.AutoField(primary_key=True)
    quote_uuid = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        primary_key=False,
        unique=True,
    )
    quote_title = models.CharField(max_length=255)
    quote_valid_till = models.DateTimeField()
    quote_stage = models.ForeignKey(
        'list_of_quote_stage',
        on_delete=models.CASCADE,
    )
    quote_billing_address=models.ForeignKey(
        'campus',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    is_invoice = models.BooleanField(
        default=False,
    )
    """
    quote_approval_status_id = models.CharField(
        max_length=10,
        choices=QUOTE_APPROVAL_STATUS,
        default='DRAFT',
    )
    """
    quote_terms = models.TextField(
        null=True,
        blank=True,
    )
    customer_notes = models.TextField(
        null=True,
        blank=True,
    )
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        db_column='project_id',
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        db_column='task_id',
        null=True,
        blank=True,
    )
    opportunity = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
        db_column='opportunity_id',
        null=True,
        blank=True,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        db_column='customer_id',
        null=True,
        blank=True,
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        db_column='organisation_id',
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.quote_title)

    class Meta:
        db_table = "quote"


class request_for_change(models.Model):
    """
    Due to the long and complicated name, request for change will be shortened to rfc for ALL fields.
    """
    rfc_id=models.AutoField(primary_key=True)
    rfc_title=models.CharField(
        max_length=255,
    )
    rfc_summary=models.TextField(
        'rfc_summary'
    )
    rfc_type=models.IntegerField(
        choices=RFC_TYPE,
    )
    rfc_implementation_start_date=models.DateTimeField()
    rfc_implementation_end_date=models.DateTimeField()
    rfc_implementation_release_date=models.DateTimeField()
    rfc_version_number=models.CharField(
        max_length=25,
        blank=True,
        null=True,
    )
    rfc_status=models.IntegerField(
        choices=RFC_STATUS,
    )
    rfc_lead = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='rfc_lead',

    )
    rfc_priority=models.IntegerField(
        choices=RFC_PRIORITY,
        default=1,
    )
    rfc_risk = models.IntegerField(
        choices=RFC_RISK,
        default=1,
    )
    rfc_impact = models.IntegerField(
        choices=RFC_IMPACT,
        default=1,
    )
    rfc_risk_and_impact_analysis=models.TextField(
        'rfc_risk_and_impact_analysis',
    )
    rfc_implementation_plan=models.TextField(
        'rfc_implementation_plan',
    )
    rfc_backout_plan=models.TextField(
        'rfc_backout_plan',
    )
    rfc_test_plan=models.TextField(
        'rfc_test_plan',
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.rfc_title)

    class Meta:
        db_table = "request_for_change"


class request_for_change_group_approval(models.Model):
    rfc_group_approval_id=models.AutoField(primary_key=True)
    rfc=models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
    )
    group=models.ForeignKey(
        'group',
        on_delete=models.CASCADE,
    )
    approval=models.IntegerField(
        choices=RFC_APPROVAL,
        default=1,  # Waiting
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.approval)

    class Meta:
        db_table = "request_for_change_group_approval"

class request_for_change_note(models.Model):
    rfc_note_id=models.AutoField(primary_key=True)
    rfc_note=models.TextField(
        blank=True,
        null=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.rfc_note)

    class Meta:
        db_table = "request_for_change_note"


class request_for_change_stakeholder(models.Model):
    """
    This model will store all the stakeholders for those request for changes. The stakeholders could be an organisation
    OR a customer.

    rfc = request for change. It is shortened to make it easier for the programmer.
    """
    rfc_stakeholder_id=models.AutoField(primary_key=True)
    request_for_change = models.ForeignKey(
        'request_for_change',
        on_delete=models.CASCADE,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "request_for_change_stakeholder"


class requirement(models.Model):
    requirement_id = models.AutoField(primary_key=True)
    requirement_title = models.CharField(
        max_length=255,
    )
    requirement_scope = models.TextField(
        null=True,
        blank=True,
    )
    requirement_type = models.ForeignKey(
        'list_of_requirement_type',
        on_delete=models.CASCADE,
    )
    requirement_status = models.ForeignKey(
        'list_of_requirement_status',
        on_delete=models.CASCADE,
    )
    requirement_story_point_min = models.IntegerField(default=1)
    requirement_story_point_max = models.IntegerField(default=4)
    organisation = models.ForeignKey(
        organisation,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_title)

    class Meta:
        db_table = "requirement"


class requirement_customer(models.Model):
    requirement_customer_id = models.AutoField(primary_key=True)
    requirement = models.ForeignKey(
        requirement,
        on_delete=models.CASCADE,
    )
    customer = models.ForeignKey(
        customer,
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_customer_id)

    class Meta:
        db_table = "requirement_customer"


class requirement_item(models.Model):
    requirement_item_id = models.AutoField(primary_key=True)
    requirement = models.ForeignKey(
        'requirement',
        on_delete=models.CASCADE,
    )
    requirement_item_title = models.CharField(max_length=255)
    requirement_item_scope = models.TextField(
        null=True,
        blank=True,
    )
    requirement_item_status = models.ForeignKey(
        'list_of_requirement_item_status',
        on_delete=models.CASCADE,
    )
    requirement_item_type = models.ForeignKey(
        'list_of_requirement_item_type',
        on_delete=models.CASCADE,
    )
    ri_story_point_min = models.IntegerField(default=4)
    ri_story_point_max = models.IntegerField(default=10)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.requirement_item_title)

    class Meta:
        db_table = "requirement_item"


class tag(models.Model):
    tag_id = models.AutoField(primary_key=True)
    tag_name = models.CharField(
        max_length=50,
        unique=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.tag_name)

    class Meta:
        db_table = "tag"


class tag_assignment(models.Model):
    tag_assignment_id = models.AutoField(primary_key=True)
    tag = models.ForeignKey(
        tag,
        on_delete=models.CASCADE,
    )
    project=models.ForeignKey(
        project,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    opportunity = models.ForeignKey(
        opportunity,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    requirement = models.ForeignKey(
        requirement,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table="tag_assignment"



class task(models.Model):
    task_id = models.AutoField(primary_key=True)
    task_short_description = models.CharField(max_length=255)
    task_long_description = models.TextField()
    organisation = models.ForeignKey(
        'organisation',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task_start_date = models.DateTimeField()
    task_end_date = models.DateTimeField()
    task_assigned_to = models.ForeignKey(
        User,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    task_status = models.CharField(
        max_length=15,
        choices=PROJECT_STATUS_CHOICE,
        default='New'
    )
    task_story_point_min = models.IntegerField(default=4)
    task_story_point_max = models.IntegerField(default=10)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    creation_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_creation_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return str(self.task_short_description)

    class Meta:
        db_table = "task"


class task_action(models.Model):
    task_action_id = models.AutoField(primary_key=True)
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
    )
    task_action = models.TextField()
    submitted_by = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "task_action"


class task_customer(models.Model):
    task_customer_id = models.AutoField(primary_key=True)
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
    )
    customer = models.ForeignKey(
        'customer',
        on_delete=models.CASCADE,
    )
    customer_description = models.CharField(
        max_length=155,
        null=True,
        blank=True
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "task_customer"

"""
class task_group(models.Model):
    task_group_id = models.AutoField(primary_key=True)
    task_id = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
    )
    group_id = models.ForeignKey(
        'group',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "task_group"
"""

"""
class task_history(models.Model):
    task_history_id = models.AutoField(primary_key=True)
    task_id = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
    )
    user_id = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )
    user_infomation = models.CharField(max_length=255)
    task_history = models.TextField()
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "task_history"
"""

"""
class task_opportunity(models.Model):
    task_opportunity_id = models.AutoField(primary_key=True)
    task_id = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
    )
    opportunity_id = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "task_opportunity"
"""

class timesheet(models.Model):
    timesheet_id = models.AutoField(primary_key=True)
    timesheet_description = models.CharField(
        max_length=255,
        blank=True,
        null=True,
    )
    timesheet_date = models.DateField()
    timesheet_start_time = models.TimeField()
    timesheet_end_time = models.TimeField()
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    requirement_item = models.ForeignKey(
        'requirement_item',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    #Doubles up as the user inputting the time
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "timesheet"

class to_do(models.Model):
    to_do_id = models.AutoField(primary_key=True)
    to_do = models.CharField(
        max_length=255,
    )
    to_do_completed = models.BooleanField(default=False)
    project = models.ForeignKey(
        'project',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    task = models.ForeignKey(
        'task',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    opportunity = models.ForeignKey(
        'opportunity',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "to_do"


class user_group(models.Model):
    user_group_id = models.AutoField(primary_key=True)
    username = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )
    group = models.ForeignKey(
        'group',
        on_delete=models.CASCADE,
    )
    permission_set = models.ForeignKey(
        'permission_set',
        on_delete=models.CASCADE,
    )
    report_to=models.ForeignKey(
        User,
        related_name='report_to',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    group_leader=models.BooleanField(
        default=False,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    class Meta:
        db_table = "user_group"



class user_want(models.Model):
    user_want_id=models.AutoField(
        primary_key=True,
    )
    want_choice=models.CharField(
        max_length=50,
        choices=WANT_CHOICE,
    )
    want_choice_text=models.CharField(
        max_length=50,
    )
    want_skill=models.CharField(
        max_length=50,
        choices=SKILL_CHOICE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return self.want_choice_text

    class Meta:
        db_table = "user_want"


class user_weblink(models.Model):
    user_weblink_id=models.AutoField(primary_key=True)
    user_weblink_url=models.URLField(max_length=255)
    user_weblink_source=models.CharField(
        max_length=50,
        choices=WEBSITE_SOURCE,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return self.user_weblinks_url

    class Meta:
        db_table = "user_weblink"


class whiteboard(models.Model):
    whiteboard_id=models.AutoField(primary_key=True)
    whiteboard_title=models.CharField(max_length=255)
    whiteboard_xml=models.TextField(
        null=True,
        blank=True,
    )
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    change_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='%(class)s_change_user'
    )
    is_deleted = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return self.whiteboard_title

    class Meta:
        db_table = "whiteboard"



# Import Forms
from ..forms import *

# Import Django Libraries
from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.template import loader
from django.urls import reverse
from django.views.decorators.http import require_http_methods
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Count

# Import Python Libraries
import json, urllib.parse, random

# Import user permission library
from ..user_permissions import return_user_permission_level


@login_required(login_url='login', redirect_field_name="")
def dashboard(request):
    """
    Due to a bug - if the user goes to /admin/ and logs in there, they will by pass this one session request. It is
    placed here to make sure. :)
    """
    request.session['is_superuser'] = request.user.is_superuser

    # Get user's default permissions
    permission_results = return_user_permission_level(request, None, 'project')

    # Load the template
    t = loader.get_template('NearBeach/dashboard/dashboard.html')

    # context
    c = {
        'new_item_permission': permission_results['new_item'],
        'administration_permission': permission_results['administration'],
    }

    return HttpResponse(t.render(c, request))


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def get_bug_list(request):
    """

    :param request:
    :return:
    """
    bug_results = bug.objects.filter(
        is_deleted=False,
        # Add in ability to tell if bugs are opened or closed
    ).values('bug_status').annotate(Count("bug_status"))

    # Send back json data
    json_results = json.dumps(list(bug_results), cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login', redirect_field_name='')
@require_http_methods(['POST'])
def get_my_objects(request):
    """
    
    :param request:
    :return:
    """

    # Get the user data
    project_results = project.objects.filter(
        is_deleted=False,
        project_id__in=object_assignment.objects.filter(
            is_deleted=False,
            project_id__isnull=False,
            assigned_user=request.user,
        ).values('project_id')
    )

    requirement_results = requirement.objects.filter(
        is_deleted=False,
        requirement_id__in=object_assignment.objects.filter(
            is_deleted=False,
            requirement_id__isnull=False,
            assigned_user=request.user,
        ).values('requirement_id')
    )

    task_results = task.objects.filter(
        is_deleted=False,
        task_id__in=object_assignment.objects.filter(
            is_deleted=False,
            task_id__isnull=False,
            assigned_user=request.user,
        ).values('task_id')
    )

    # Only have 25 results and order by alphabetical order
    # requirement_results.order_by('requirement_title')[:25]
    # project_results.order_by('project_name')[:25]
    # task_results.order_by('task_short_description').values()[:25]

    """
    The pain point
    ~~~~~~~~~~~~~~
    Due to Django wanting to send converted json data as a string, we have to;
    1. Apply serialisation
    2. Apply a json.loads function
    3. Compile data and send back.
    
    Note to Django developers - there has to be a better way
    """
    requirement_results = serializers.serialize('json', requirement_results)
    # requirement_results = json.dumps(list(requirement_results), cls=DjangoJSONEncoder)
    project_results = serializers.serialize('json', project_results)
    task_results = serializers.serialize('json', task_results)

    # Send back a JSON array with JSON arrays inside
    return JsonResponse({
        'requirement': json.loads(requirement_results),
        'project': json.loads(project_results),
        'task': json.loads(task_results),
    })


@login_required(login_url='login', redirect_field_name='')
@require_http_methods(['POST'])
def rfc_approvals(request):
    """

    :param request:
    :return:
    """

    # Get a list of RFC's that are awaiting approval
    rfc_results = request_for_change.objects.filter(
        is_deleted=False,
        rfc_status=2,  # Waiting for approval
    )

    """
    Filter the rfc_results, with any object assignment that the user is currently;
    - A group leader of
    - Connected to an RFC currently waiting for approval
    """
    rfc_results = rfc_results.filter(
        # Filter for any object assignment that is connected by group that user is group leader of.
        rfc_id__in=object_assignment.objects.filter(
            is_deleted=False,
            request_for_change_id__in=rfc_results.values('rfc_id'),
            group_id__in=user_group.objects.filter(
                is_deleted=False,
                username_id=request.user,
                group_leader=True,
            ).values('group_id'),
        ).values('request_for_change_id')
    )

    return HttpResponse(serializers.serialize('json',rfc_results), content_type='application/json')




"""
The following code has been sourced from DJANGO-PRIVATE-MEDIA - https://github.com/RacingTadpole/django-private-media
which has been sourced from django-filer.

I have modified some of the code, however most of the recognition should be for both;
-- django-filer
-- django-private-media

I have placed this code into one file for convenience.
"""

from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.http import Http404, HttpResponseNotModified, HttpResponse
from django.utils.http import http_date
from django.views.static import was_modified_since

import mimetypes, os, stat

"""
FUNCTION DESCRIPTION
~~~~~~~~~~~~~~~~~~~~
To return the location of where we are going to store the file. It will
look through the SETTINGS.PY file for the Django project and extract the
variables;
-- PRIVATE_MEDIA_ROOT
-- PRIVATE_MEDIA_URL

These will need to be set in the SETTINGS.PY
"""


class File_Storage(FileSystemStorage):
    def __init__(self, location=None, base_url=None):
        if location is None:
            location = settings.PRIVATE_MEDIA_ROOT
        if base_url is None:
            base_url = settings.PRIVATE_MEDIA_URL
        return super(File_Storage, self).__init__(location, base_url)


"""
FUNCTION DESCRIPTION
~~~~~~~~~~~~~~~~~~~~
To return a TRUE or FALSE result depending if the user has permission
to access the file.

METHOD
~~~~~~
1.) Check to make sure a user is logged in
2.) Download the permission rows for the file
3.) Determine if the user meets either 1 of the 
    following permissions;@login_required(login_url='login')
    -- No restrictions on file
    -- User is in group allocated permission to file
    -- User has been allocated permission
    If user meets at least one condition, then the value
    TRUE is returned, else FALSE
"""


class Check_Permissions(object):
    def has_read_permission(self, request, path):
        # Is user logged in
        user = request.user
        if not user.is_authenticated():
            return False

        # TEMP RETURN LINE
        return True

        """
        The following code can expandable for all different modules that require document
        uploads. It will need to be tested with a very large dataset.
        """
        current_user = request.user

        # Setup connection to the database and query it
        cursor = connection.cursor()

        cursor.execute("""
        SELECT 
        count(project_group.groups_id_id)
        + CASE WHEN document_permission.organisations_id_id IS NULL THEN 0 ELSE 1 END
        + CASE WHEN document_permission.customer_id_id IS NULL THEN 0 ELSE 1 END
        AS PERMISSION

        FROM
        user_group
        , document_permission
            left join project_group
              on document_permission.project_id_id = project_group.project_id_id
              AND project_group.groups_id_id = user_group.group_id_id
            left join tasks_group
              on document_permission.task_id_id = tasks_group.tasks_id_id
              AND tasks_group.groups_id_id = user_group.group_id_id
        /*
        MISSING
        ~~~~~~~
        Currently opportunities do not have a limit to. I will need to
        apply this.
        */


        WHERE 1=1
        -- THE USER INPUT
        AND user_group.username_id = %s

        --THE DOCUMENT INPUT
        AND document_permission.document_key_id = %s

        -- Make sure the doc is not deleted
        AND document_permission.is_deleted = 'FALSE'
       	""", [current_user.id, object])
        has_permission = cursor.fetchall()

        if not has_permission[0][0] >= 1:
            # User have permission to view the file
            return True
        else:
            return False


"""
Left unchanged
"""


class NginxXAccelRedirectServer(object):
    def serve(self, request, path):
        response = HttpResponse()
        fullpath = os.path.join(settings.PRIVATE_MEDIA_ROOT, path)
        response['X-Accel-Redirect'] = fullpath
        response['Content-Type'] = mimetypes.guess_type(path)[0] or 'application/octet-stream'
        return response


"""
Left unchanged
"""


class ApacheXSendfileServer(object):
    def serve(self, request, path):
        fullpath = os.path.join(settings.PRIVATE_MEDIA_ROOT, path)
        response = HttpResponse()
        response['X-Sendfile'] = fullpath

        # From django-filer (https://github.com/stefanfoulis/django-filer/):
        # This is needed for lighttpd, hopefully this will
        # not be needed after this is fixed:
        # http://redmine.lighttpd.net/issues/2076
        response['Content-Type'] = mimetypes.guess_type(path)[0] or 'application/octet-stream'

        # filename = os.path.basename(path)
        # response['Content-Disposition'] = smart_str(u'attachment; filename={0}'.format(filename))

        return response


"""
Left unchanged
"""


class DefaultServer(object):
    """
    Serve static files from the local filesystem through django.
    This is a bad idea for most situations other than testing.

    This will only work for files that can be accessed in the local filesystem.
    """

    def serve(self, request, path):
        # the following code is largely borrowed from `django.views.static.serve`
        # and django-filetransfers: filetransfers.backends.default
        fullpath = os.path.join(settings.PRIVATE_MEDIA_ROOT, path)
        if not os.path.exists(fullpath):
            raise Http404('"{0}" does not exist'.format(fullpath))
        # Respect the If-Modified-Since header.
        statobj = os.stat(fullpath)
        content_type = mimetypes.guess_type(fullpath)[0] or 'application/octet-stream'
        if not was_modified_since(request.META.get('HTTP_IF_MODIFIED_SINCE'),
                                  statobj[stat.ST_MTIME], statobj[stat.ST_SIZE]):
            return HttpResponseNotModified(content_type=content_type)
        response = HttpResponse(open(fullpath, 'rb').read(), content_type=content_type)
        response["Last-Modified"] = http_date(statobj[stat.ST_MTIME])
        # filename = os.path.basename(path)
        # response['Content-Disposition'] = smart_str(u'attachment; filename={0}'.format(filename))
        return response


"""
Left unchanged
"""
from importlib import import_module


def get_class(import_path=None):
    """
    Largely based on django.core.files.storage's get_storage_class
    """
    from django.core.exceptions import ImproperlyConfigured
    if import_path is None:
        raise ImproperlyConfigured('No class path specified.')
    try:
        dot = import_path.rindex('.')
    except ValueError:
        raise ImproperlyConfigured("%s isn't a module." % import_path)
    module, classname = import_path[:dot], import_path[dot + 1:]
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured('Error importing module %s: "%s"' % (module, e))
    try:
        return getattr(mod, classname)
    except AttributeError:
        raise ImproperlyConfigured('Module "%s" does not define a "%s" class.' % (module, classname))


"""
Mostly left unchanged
"""
if settings.DEBUG == True:
    server = DefaultServer(**getattr(settings, 'PRIVATE_MEDIA_SERVER_OPTIONS', {}))
else:
    server = ApacheXSendfileServer(**getattr(settings, 'PRIVATE_MEDIA_SERVER_OPTIONS', {}))
if hasattr(settings, 'PRIVATE_MEDIA_PERMISSIONS'):
    permissions = get_class(settings.PRIVATE_MEDIA_PERMISSIONS)(
        **getattr(settings, 'PRIVATE_MEDIA_PERMISSIONS_OPTIONS', {}))
else:
    permissions = Check_Permissions()




from django.http import HttpResponse
from django.template import loader
from django.contrib.auth.decorators import login_required


@login_required(login_url='login',redirect_field_name="")
def error_403(request, exception):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    t = loader.get_template('403.html')

    c = {}

    return HttpResponse(t.render(c, request))


@login_required(login_url='login',redirect_field_name="")
def error_404(request, exception):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    t = loader.get_template('404.html')

    c = {}

    return HttpResponse(t.render(c, request))


@login_required(login_url='login',redirect_field_name="")
def error_500(request):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    t = loader.get_template('500.html')

    c = {}

    return HttpResponse(t.render(c, request))


from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min
from NearBeach.forms import *
from NearBeach.user_permissions import return_user_permission_level
from django.views.decorators.http import require_http_methods
from django.shortcuts import get_object_or_404


import json


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_requirement_link(request,requirement_id):
    # ADD IN PERMISSION
    permission_results = return_user_permission_level(request, None, ['requirement','requirement_link'])
    # What about requirement items? Will need to fix this elegantly.

    if permission_results['requirement'] < 2:
        return HttpResponseRedirect(reverse('permission_denied'))

    form = AddRequirementLinkForm(request.POST)

    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)


    # Get the requirement instnace
    requirement_instance = requirement.objects.get(requirement_id=requirement_id)

    # Get the project list from the form
    for row in request.POST.getlist("project"):
        submit_object_assignment = object_assignment(
            requirement=requirement_instance,
            project=project.objects.get(project_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    for row in request.POST.getlist("task"):
        submit_object_assignment = object_assignment(
            requirement=requirement_instance,
            task=task.objects.get(task_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    for row in request.POST.getlist("opportunity"):
        submit_object_assignment = object_assignment(
            requirement=requirement_instance,
            opportunity=opportunity.objects.get(opportunity_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    return HttpResponse("Success")


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def get_requirement_item_links(request,requirement_id):
    # Get the requirement information
    requirement_results = requirement.objects.get(requirement_id=requirement_id)

    # Check the permissions
    permission_results = get_user_requirement_permissions(request, requirement_id)

    # If user has no permissions to this requirement send them to the appropriate location
    if permission_results['requirement'] == 0:
        # Users who create the requirement get at least read only
        if requirement_results.creation_user == request.user:
            return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_id}))

        # Users who did not create the requirement get sent to permission denied.
        return HttpResponseRedirect(reverse('permission_denied'))

    # Use object_assignment to get the requirme
    link_results = object_assignment.objects.filter(
        Q(
            is_deleted=False,
            requirement_item_id__in=requirement_item.objects.filter(
                is_deleted=False,
                requirement_id=requirement_id,
            ).values('requirement_item_id')
        ) & Q(
            Q(opportunity_id__isnull=False) |
            Q(quote_id__isnull=False) |
            Q(project_id__isnull=False) |
            Q(task_id__isnull=False)
        )
    ).values(
        'opportunity_id',
        'opportunity_id__opportunity_name',
        'opportunity_id__opportunity_stage_id__opportunity_stage_description',
        'quote_id',
        'quote_id__quote_title',
        'quote_id__quote_stage_id__quote_stage',
        'project_id',
        'project_id__project_name',
        'project_id__project_status',
        'task_id',
        'task_id__task_short_description',
        'task_id__task_status',
        'requirement_item_id',
        'requirement_item_id__requirement_item_title',
    )

    """
    As explained on stack overflow here - https://stackoverflow.com/questions/7650448/django-serialize-queryset-values-into-json#31994176
    We need to Django's serializers can't handle a ValuesQuerySet. However, you can serialize by using a standard 
    json.dumps() and transforming your ValuesQuerySet to a list by using list().[sic]
    """

    # Send back json data
    json_results = json.dumps(list(link_results), cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')

@login_required(login_url='login',redirect_field_name="")
def get_requirement_item_status_list(request,requirement_id):
    # Get all status - even deleted ones.
    status_list = list_of_requirement_item_status.objects.all()

    # Send back json data
    json_results = serializers.serialize('json', status_list)

    return HttpResponse(json_results, content_type='application/json')

@login_required(login_url='login',redirect_field_name="")
def get_requirement_item_type_list(request,requirement_id):
    # Get all status - even deleted ones.
    type_list = list_of_requirement_item_type.objects.all()

    # Send back json data
    json_results = serializers.serialize('json', type_list)

    return HttpResponse(json_results, content_type='application/json')

@login_required(login_url='login',redirect_field_name="")
def get_requirement_items(request,requirement_id):
    # Get all the requirement items assigned to the requirement
    requirement_item_results = requirement_item.objects.filter(
        is_deleted=False,
        requirement_id=requirement_id,
    )

    # Send back json data
    json_results = serializers.serialize('json', requirement_item_results)

    return HttpResponse(json_results, content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def get_requirement_links_list(request,requirement_id):
    # Get the requirement information
    requirement_results = requirement.objects.get(requirement_id=requirement_id)

    # Check the permissions
    permission_results = get_user_requirement_permissions(request,requirement_id)

    # If user has no permissions to this requirement send them to the appropriate location
    if permission_results['requirement'] == 0:
        # Users who create the requirement get at least read only
        if requirement_results.creation_user == request.user:
            return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_id}))

        # Users who did not create the requirement get sent to permission denied.
        return HttpResponseRedirect(reverse('permission_denied'))

    # Use object_assignment to get the requirme
    link_results = object_assignment.objects.filter(
        Q(
            is_deleted=False,
            requirement_id=requirement_id,
        ) & Q(
            Q(quote_id__isnull=False) |
            Q(project_id__isnull=False) |
            Q(task_id__isnull=False)
        )
    ).values(
        'project_id',
        'project_id__project_name',
        'project_id__project_status',
        'task_id',
        'task_id__task_short_description',
        'task_id__task_status',
        'requirement_item_id',
        'requirement_item_id__requirement_item_title',
    )

    """
    As explained on stack overflow here - https://stackoverflow.com/questions/7650448/django-serialize-queryset-values-into-json#31994176
    We need to Django's serializers can't handle a ValuesQuerySet. However, you can serialize by using a standard 
    json.dumps() and transforming your ValuesQuerySet to a list by using list().[sic]
    """

    # Send back json data
    json_results = json.dumps(list(link_results),cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def get_user_requirement_permissions(request,requirement_id):
    """
    Use the requirement_id and find out if the user has access to this requirement
    :param requirement_id:
    :return:
    """
    requirement_groups = object_assignment.objects.filter(
        is_deleted=False,
        #requirement_id=requirement_id
    ).values('group_id')

    if requirement_id > 0:
        # Make sure to filter by requirement groups
        requirement_groups = requirement_groups.filter(
            requirement_id=requirement_id
        )

    return return_user_permission_level(request, requirement_groups, ['requirement','requirement_link'])


@login_required(login_url='login',redirect_field_name="")
def new_requirement(request, location_id="", destination=""):
    """
    Loads the new requirement page
    :param request:
    :param location_id:
    :param destination:
    :return:
    """
    # Get user permission
    permission_results = get_user_requirement_permissions(request,0)

    # If user has no permissions to create requirements, then send them to the appropriate location
    if permission_results['requirement'] <= 2:
        # Users can not create requirement.
        return HttpResponseRedirect(reverse('permission_denied'))

    #Extract Data
    status_list = list_of_requirement_status.objects.filter(
        is_deleted=False,
        requirement_status_is_closed=False,
    )

    type_list = list_of_requirement_type.objects.filter(
        is_deleted=False,
    )

    group_results = group.objects.filter(
        is_deleted=False,
    )

    #Load template
    t = loader.get_template('NearBeach/requirements/new_requirements.html')

    # context
    c = {
        'status_list': serializers.serialize("json",status_list),
        'type_list': serializers.serialize("json",type_list),
        'group_results': serializers.serialize("json",group_results),
    }

    return HttpResponse(t.render(c, request))

@login_required(login_url='login',redirect_field_name='')
def new_requirement_save(request, location_id="", destination=""):
    # Make sure this is a post
    if not request.method == "POST":
        #Give the user a 404
        return HttpResponseBadRequest("Sorry - Post only")

    # Get user permission
    permission_results = get_user_requirement_permissions(request,0)

    # If user has no permissions to create requirements, then send them to the appropriate location
    if permission_results['requirement'] <= 2:
        # Users can not create requirement.
        return HttpResponseRedirect(reverse('permission_denied'))

    # Get the data and place into the form

    form = NewRequirementForm(request.POST)

    if not form.is_valid():
        # Something went wrong with the form.
        return HttpResponseBadRequest("There was something wrong with the form")

    # Save the form
    submit_requirement = requirement(
        requirement_title=form.cleaned_data['requirement_title'],
        requirement_scope=form.cleaned_data['requirement_scope'],
        organisation=form.cleaned_data['organisation'],
        requirement_status=form.cleaned_data['requirement_status'],
        requirement_type=form.cleaned_data['requirement_type'],
        change_user=request.user,
        creation_user=request.user,
    )
    submit_requirement.save()

    # Get the group list and apply the permissions
    group_list = request.POST.getlist("group_list")

    for single_group in group_list:
        # Get the group instance
        group_instance = group.objects.get(group_id=single_group)

        # Save the group instance against object assignment
        submit_object_assignment = object_assignment(
            group_id=group_instance,
            requirement=submit_requirement,
            change_user=request.user,
        )

        # Save
        submit_object_assignment.save()

    # Send back requirement_information URL
    return HttpResponse(reverse('requirement_information',args={submit_requirement.requirement_id}))



@login_required(login_url='login',redirect_field_name='')
def requirement_information(request, requirement_id):
    """
    Loads the requirement information.
    :param request:
    :param requirement_id:
    :return:
    """
    # Get the requirement information
    # requirement_results = requirement.objects.get(requirement_id=requirement_id)
    requirement_results = get_object_or_404(requirement,requirement_id=requirement_id)

    # Check the permissions
    permission_results = get_user_requirement_permissions(request,requirement_id)

    # If user has no permissions to this requirement send them to the appropriate location
    if permission_results['requirement'] == 0:
        # Users who create the requirement get at least read only
        if requirement_results.creation_user == request.user:
            return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_id}))

        # Users who did not create the requirement get sent to permission denied.
        return HttpResponseRedirect(reverse('permission_denied'))


    # If the requirement has been closed - send user to the read only section
    if requirement_results.requirement_status.requirement_status == "Completed":
        return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_id}))


    #Load template
    t = loader.get_template('NearBeach/requirements/requirement_information.html')

    #Get any extra data required
    organisation_results = organisation.objects.get(
        organisation_id=requirement_results.organisation_id,
    )

    status_list = list_of_requirement_status.objects.filter(
        is_deleted=False,
        requirement_status_is_closed=False,
    )

    type_list = list_of_requirement_type.objects.filter(
        is_deleted=False,
    )

    group_results = group.objects.filter(
        is_deleted=False,
    )

    requirement_item_results = requirement_item.objects.filter(
        is_deleted=False,
        requirement_id=requirement_id,
    )

    # context
    c = {
        'group_results': serializers.serialize("json", group_results),
        'organisation_results': serializers.serialize("json", [organisation_results]),
        'permission_results': permission_results,
        'requirement_results': serializers.serialize("json", [requirement_results]),
        'requirement_id': requirement_id,
        'requirement_item_results': serializers.serialize("json",requirement_item_results),
        'status_list': serializers.serialize("json", status_list),
        'type_list': serializers.serialize("json", type_list),

    }

    return HttpResponse(t.render(c, request))

@login_required(login_url='login',redirect_field_name="")
def requirement_information_save(request, requirement_id):
    """

    :param request:
    :param requirement_id:
    :return:
    """
    # This request needs to be in POST
    if request.method != "POST":
        return HttpResponseBadRequest("Request must be in POST")

    # Check the permissions
    permission_results = get_user_requirement_permissions(request, requirement_id)

    # Users will need an edit permission
    if permission_results['requirement'] <= 1:
        # On your bike - you do not have enough permission
        return HttpResponseRedirect(reverse('permission_denied'))

    # Insert POST into form
    form = UpdateRequirementForm(request.POST)

    # If there is an error - notify the user
    if not form.is_valid():
        return HttpResponseBadRequest("Sorry, there is an error with the form: %s" % form.errors)


    # Get the requirement
    requirement_result = requirement.objects.get(requirement_id=requirement_id)

    # Update all the fields
    requirement_result.requirement_title = form.cleaned_data['requirement_title']
    requirement_result.requirement_scope = form.cleaned_data['requirement_scope']
    requirement_result.requirement_status = form.cleaned_data['requirement_status']
    requirement_result.requirement_type = form.cleaned_data['requirement_type']

    requirement_result.save()

    # Return a success
    return HttpResponse("Requirement Saved")


@login_required(login_url='login',redirect_field_name="")
def update_requirement(request,requirement_id):
    return False


# Generated by Django 3.1.2 on 2021-01-23 04:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('NearBeach', '0004_auto_20201114_2110'),
    ]

    operations = [
        migrations.AddField(
            model_name='change_task',
            name='is_downtime',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='change_task',
            name='change_task_description',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name='change_task_block',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('is_deleted', models.BooleanField(default=False)),
                ('blocked_change_task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blocked_change_task', to='NearBeach.change_task')),
                ('change_task_blocks', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='change_task_blocks', to='NearBeach.change_task')),
                ('change_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='change_task_block_change_user', to=settings.AUTH_USER_MODEL)),
                ('creation_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='change_task_block_creation_user', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'change_task_block',
            },
        ),
    ]



from django.test import TestCase, Client
from django.urls import reverse
from .function_tests import *

"""
The following test will;
1. Not log in a user
2. Visit every single URL

Where the expected results would be;
1. Login page will load correctly
2. Every other page redirects to the login page
"""


class CheckLoginPage(TestCase):
    def test_login_page(self):
        # Make sure the login page does work
        c = Client()
        response = c.get(reverse('login'))
        self.assertEqual(response.status_code, 200)
        print("Non logged in users can get access to the login page")


class CheckCustomerInformation(TestCase):
    def test_customer_information_page(self):
        # Make sure the customer information page redirects
        c = Client()
        response_get = c.get(reverse('customer_information', args=[1]))
        response_post = c.post(reverse('customer_information_save',args=[1]))

        # Check
        self.assertRedirects(
            response_get,
            reverse('login'),
            status_code=302,
            target_status_code=200,
            msg_prefix='',
            fetch_redirect_response=True
        )

        self.assertRedirects(
            response_post,
            reverse('login'),
            status_code=302,
            target_status_code=200,
            msg_prefix='',
            fetch_redirect_response=True
        )

        # Notify tester
        print("Non logged in users get redirected when visiting customers")


class CheckDashboard(TestCase):
    def test_dashboard(self):
        # Make sure the user gets redirected to the login page
        c = Client()

        # Setup the response array
        response_array = [
            c.get(reverse('dashboard')),
            c.get(reverse('get_bug_list')),
            c.get(reverse('get_my_objects')),
        ]

        # Check the data in the array
        assertRedirectsToLogin(response_array, self)

        # Tell tester of results
        print("Non Logged in users redirected when visiting dashboard")


class CheckDocumentation(TestCase):
    def test_documentation(self):
        # Make sure the user gets redirected to the login page
        c = Client()

        # The get components should receive a 405
        response_get_array = [
            c.get(reverse('document_add_folder', args=['project', 1])),
            c.get(reverse('document_add_link', args=['project', 1])),
            c.get(reverse('document_list_files', args=['project', 1])),
            c.get(reverse('document_list_folders', args=['project', 1])),
            c.get(reverse('document_upload', args=['project', 1])),
            c.get(reverse('document_get_max_upload')),
        ]

        # Check the array
        assertEqual405(response_get_array, self)

        # The POST components should be redirected
        response_post_array = [
            c.post(reverse('document_add_folder', args=['project', 1])),
            c.post(reverse('document_add_link', args=['project', 1])),
            c.post(reverse('document_list_files', args=['project', 1])),
            c.post(reverse('document_list_folders', args=['project', 1])),
            c.post(reverse('document_upload', args=['project', 1])),
            c.post(reverse('document_get_max_upload')),
        ]

        #Check the array
        assertRedirectsToLogin(response_post_array, self)

        print("Non Logged in users redirected when visiting documentation")


class CheckKanban(TestCase):
    def test_kanban_information(self):
        # Make sure the user gets redirected to the login page
        c = Client()

        # Setup response array
        response_array = [
            c.get(reverse('kanban_information', args=[1])),
            c.get(reverse('add_kanban_link', args=[1, 'project'])),
            c.get(reverse('kanban_link_list', args=[1, 'project'])),
            c.get(reverse('new_kanban_card', args=[1])),
            c.get(reverse('move_kanban_card', args=[1])),
            c.get(reverse('check_kanban_board_name')),
            c.get(reverse('kanban_update_card')),
        ]

        # Check the array
        assertRedirectsToLogin(response_array, self)

        print("Non Logged in users redirected when visiting kanban information")


class CheckPrivateDocument(TestCase):
    def test_private_document(self):
        # Make sure the user gets redirected to the login page
        c = Client()
        response_get = c.get(reverse('private_download_file', args=['12345678-1234-5678-1234-567812345678']))

        self.assertRedirects(
            response_get,
            reverse('login'),
            status_code=302,
            target_status_code=200,
            msg_prefix='',
            fetch_redirect_response=True
        )

        print("Non Logged in users redirected when visiting private document")


class CheckNew(TestCase):
    def new(self):
        # Make sure the user gets redirected to the login page
        c = Client()

        response_array = [
            c.get(reverse('new_customer')),
            c.get(reverse('new_kanban')),
            c.get(reverse('new_organisation')),
            c.get(reverse('new_project')),
            c.get(reverse('new_request_for_change')),
            c.get(reverse('new_requirement')),
            c.get(reverse('new_requirement_item', args=[1])),
            c.get(reverse('new_task')),
        ]

        # Check the array
        assertEqual405(response_array, self)

        print("Non Logged in users redirected when visiting private document")

    def new_save(self):
        # Make sure the user gets redirected to the login page
        c = Client()

        response_array = [
            c.post(reverse('new_customer_save')),
            c.post(reverse('new_kanban_save')),
            c.post(reverse('new_organisation_save')),
            c.post(reverse('new_project_save')),
            c.post(reverse('new_request_for_change_save')),
            c.post(reverse('new_requirement_save')),
            c.post(reverse('new_requirement_item_save', args=[1])),
            c.post(reverse('new_task_save')),
        ]

        # Check the array
        assertRedirectsToLogin(response_array, self)


class CheckObjects(TestCase):
    def test_objects_get(self):
        # Setup the client
        c = Client()

        # Setup the request array
        response_array = [
            c.get(reverse('add_bug', args=['project',1])),
            c.get(reverse('add_customer', args=['project',1])),
            c.get(reverse('add_group', args=['project',1])),
            c.get(reverse('add_link', args=['project',1])),
            c.get(reverse('add_notes', args=['project',1])),
            c.get(reverse('add_user', args=['project',1])),
            c.get(reverse('associated_objects', args=['project',1])),
            c.get(reverse('bug_client_list')),
            c.get(reverse('bug_list', args=['project',1])),
            c.get(reverse('customer_list', args=['project',1])),
            c.get(reverse('customer_list_all', args=['project',1])),
            c.get(reverse('group_list', args=['project',1])),
            c.get(reverse('group_list_all', args=['project',1])),
            c.get(reverse('link_list', args=['project',1,'task'])),
            c.get(reverse('note_list', args=['project',1])),
            c.get(reverse('object_link_list', args=['project',1])),
            c.get(reverse('query_bug_client', args=['project',1])),
            c.get(reverse('user_list', args=['project',1])),
            c.get(reverse('user_list_all', args=['project',1])),
            c.get(reverse('lead_user_list')),
        ]

        # Cheeck the array
        assertEqual405(response_array, self)

        print("Non Logged in users redirected when visiting objects (GET)")

    def test_objects_post(self):
        # Setup the client
        c = Client()

        # Setup the request array
        response_array = [
            c.post(reverse('add_bug', args=['project', 1])),
            c.post(reverse('add_customer', args=['project', 1])),
            c.post(reverse('add_group', args=['project', 1])),
            c.post(reverse('add_link', args=['project', 1])),
            c.post(reverse('add_notes', args=['project', 1])),
            c.post(reverse('add_user', args=['project', 1])),
            c.post(reverse('associated_objects', args=['project', 1])),
            c.post(reverse('bug_client_list')),
            c.post(reverse('bug_list', args=['project', 1])),
            c.post(reverse('customer_list', args=['project', 1])),
            c.post(reverse('customer_list_all', args=['project', 1])),
            c.post(reverse('group_list', args=['project', 1])),
            c.post(reverse('group_list_all', args=['project', 1])),
            c.post(reverse('link_list', args=['project', 1, 'task'])),
            c.post(reverse('note_list', args=['project', 1])),
            c.post(reverse('object_link_list', args=['project', 1])),
            c.post(reverse('query_bug_client', args=['project', 1])),
            c.post(reverse('user_list', args=['project', 1])),
            c.post(reverse('user_list_all', args=['project', 1])),
            c.post(reverse('lead_user_list')),
        ]

        # Cheeck the array
        assertRedirectsToLogin(response_array, self)

        print("Non Logged in users redirected when visiting objects (GET)")


from django.contrib import admin

# Register your models here.
from .models import *

admin.site.register(about_user)
admin.site.register(bug)
admin.site.register(bug_client)
admin.site.register(campus)
admin.site.register(change_task)
admin.site.register(change_task_block)
admin.site.register(contact_history)
admin.site.register(cost)
admin.site.register(customer)
admin.site.register(customer_campus)
admin.site.register(document)
admin.site.register(document_permission)
admin.site.register(email_content)
admin.site.register(email_contact)
admin.site.register(folder)
admin.site.register(group_permission)
admin.site.register(group)
admin.site.register(kanban_board)
admin.site.register(kanban_card)
admin.site.register(kanban_column)
admin.site.register(kanban_level)
admin.site.register(kudos)
admin.site.register(list_of_amount_type)
admin.site.register(list_of_bug_client)
admin.site.register(list_of_currency)
admin.site.register(list_of_contact_type)
admin.site.register(list_of_country_region)
admin.site.register(list_of_country)
admin.site.register(list_of_lead_source)
admin.site.register(list_of_opportunity_stage)
admin.site.register(list_of_quote_stage)
admin.site.register(list_of_requirement_item_status)
admin.site.register(list_of_requirement_item_type)
admin.site.register(list_of_requirement_status)
admin.site.register(list_of_requirement_type)
admin.site.register(list_of_tax)
admin.site.register(list_of_title)
admin.site.register(nearbeach_option)
admin.site.register(object_assignment)
admin.site.register(object_note)
admin.site.register(opportunity)
#admin.site.register(opportunity_connection)
admin.site.register(organisation)
admin.site.register(permission_set)
admin.site.register(project_customer)
#admin.site.register(project_history)
#admin.site.register(project_stage)
#admin.site.register(project_task)
admin.site.register(project)
admin.site.register(quote)
admin.site.register(request_for_change)
admin.site.register(request_for_change_group_approval)
admin.site.register(request_for_change_stakeholder)
admin.site.register(requirement)
admin.site.register(requirement_customer)
admin.site.register(requirement_item)
#admin.site.register(stage)
admin.site.register(tag)
admin.site.register(tag_assignment)
admin.site.register(task_action)
admin.site.register(task_customer)
#admin.site.register(task_history)
admin.site.register(task)
admin.site.register(timesheet)
admin.site.register(to_do)
admin.site.register(user_group)
admin.site.register(user_want)
admin.site.register(user_weblink)
admin.site.register(whiteboard)









from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min
from NearBeach.forms import *
from NearBeach.user_permissions import return_user_permission_level
from django.views.decorators.http import require_http_methods


import json


@login_required(login_url='login',redirect_field_name="")
def new_organisation(request):
    """

    :param request:
    :return:
    """
    # Get user permission

    # Get templates
    t = loader.get_template('NearBeach/organisations/new_organisations.html')

    # Get Context
    c = {}

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name='')
def new_organisation_save(request):
    """

    :param request:
    :return:
    """
    permission_results = return_user_permission_level(request, None, 'organisation')

    if permission_results['organisation'] < 3:
        return HttpResponseRedirect(reverse('permission_denied'))

    # Get the data
    form = OrganisationForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    organisation_submit = organisation(
        change_user=request.user,
        organisation_name=form.cleaned_data['organisation_name'],
        organisation_email=form.cleaned_data['organisation_email'],
        organisation_website=form.cleaned_data['organisation_website'],
    )
    organisation_submit.save()

    return HttpResponse(reverse('organisation_information',args={organisation_submit.organisation_id}))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def organisation_duplicates(request):
    """

    :param request:
    :return:
    """

    # ADD IN USER PERMISSION CHECKS

    # Extract data from POST
    form = OrganisationForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Check to see if there are any matches
    organisation_results = organisation.objects.filter(
        Q(
            is_deleted=False,
        ) & Q(
            Q(organisation_name__contains=form.cleaned_data['organisation_name']) |
            Q(organisation_website__contains=form.cleaned_data['organisation_website']) |
            Q(organisation_email__contains=form.cleaned_data['organisation_email'])
        )
    )

    return HttpResponse(serializers.serialize('json',organisation_results),content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def organisation_information(request,organisation_id):
    """

    :param request:
    :param organisation_id:
    :return:
    """
    organisation_results = organisation.objects.get(organisation_id=organisation_id)

    customer_results = customer.objects.filter(
        is_deleted=False,
        organisation_id=organisation_id,
    )

    title_list = list_of_title.objects.filter(
        is_deleted=False,
    )

    t = loader.get_template('NearBeach/organisations/organisation_information.html')

    c = {
        'customer_results': serializers.serialize('json',customer_results),
        'organisation_id': organisation_id,
        'organisation_results': serializers.serialize('json',[organisation_results]),
        'title_list': serializers.serialize('json',title_list),
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def organisation_information_save(request,organisation_id):
    """

    :param request:
    :param organisation_id:
    :return:
    """

    # ADD IN PERMISSION CHECKING

    # Get the form data
    form = OrganisationForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the instance
    organisation_instance = organisation.objects.get(organisation_id=organisation_id)
    organisation_instance.organisation_name = form.cleaned_data['organisation_name']
    organisation_instance.organisation_email = form.cleaned_data['organisation_email']
    organisation_instance.organisation_website = form.cleaned_data['organisation_website']
    organisation_instance.save()

    return HttpResponse('')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def organisation_update_profile(request,organisation_id):
    """

    :param request:
    :param organisation_id:
    :return:
    """
    return HttpResponse('')


from django.http import HttpResponse, HttpResponseBadRequest
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods

from NearBeach.forms import AdminAddUserForm

from NearBeach.models import *


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_user(request):
    """
https://irisnx.us/
    :param request:
    :return:
    """

    # Add in user permissions

    # Check the form data
    form = AdminAddUserForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    """
    We can assume that there are;
    - Single User
    - Multiple Groups
    - Multiple Permission Sets
    
    We need to loop through these permutations, and add them into the database
    """
    # print("Permission Sets %s" % form.cleaned_data['permission_set'])

    group_results = form.cleaned_data['group']
    permission_set_results = form.cleaned_data['permission_set']
    user_results = form.cleaned_data['username']

    # Simple double for loops
    for single_group in group_results:
        for single_permission_set in permission_set_results:
            submit_user = user_group(
                username=user_results,
                permission_set=single_permission_set,
                group=single_group,
                change_user=request.user,
            )
            submit_user.save()

    return HttpResponse("")


# Generated by Django 3.1.2 on 2020-11-14 10:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NearBeach', '0003_auto_20201024_1219'),
    ]

    operations = [
        migrations.AddField(
            model_name='object_note',
            name='kanban_card',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NearBeach.kanban_card'),
        ),
        migrations.DeleteModel(
            name='kanban_comment',
        ),
    ]



"""
This python script will return the user's permission level for ANY given permission
"""
from .models import *
from django.db.models import Max

"""
Permission table
0 - No permission
1 - Read only
2 - Edit permission
3 - Create permission
4 - Admin/Delete permission
"""
def return_user_permission_level(request, group_list,permission_field):
    """

    :param request:
    :param group: limits data to a certain group - Null if no group
    :param permission_field: which permission field we will be looking at. The available list is;
        permission_set_id
        permission_set_name
        administration_assign_users_to_group
        administration_create_group
        administration_create_permission_sets
        administration_create_users
        assign_campus_to_customer
        associate_project_and_tasks
        customer
        invoice
        invoice_product
        opportunity
        organisation
        organisation_campus
        project
        quote
        request_for_change
        requirement
        requirement_link
        task
        document
        contact_history
        project_history
        task_history
        whiteboard

        Please note - if you want to look up more than ONE permission, please include them in [] brackets. For example if
        you would like to look up; project, project_history, and document, then you would use ['project','project_history','document']
    :param min_permission_level: tells us what is the minimum level the user has to be, if they do not meet this requirement
        then the system will formward them onto the permission denied page. Default is 1 (read only)
    :return:
    """

    #Make sure the permission_field is an array/list
    if not isinstance(permission_field, list):
        permission_field = [permission_field]

    #Default NO PERMISSION
    user_permission_level = {}

    """
    If the user is a superuser, we will return 4 no matter what. This part of the script will see if the user is actually
    a super user. If they are then it will return 4 for everything.
    """
    if request.user.is_superuser == True:
        #Add 4 to all permissions
        for row in permission_field:
            user_permission_level[row] = 4
        #User can add new items and do administration
        user_permission_level['new_item'] = 4
        user_permission_level['administration'] = 4
        return user_permission_level

    """
    We are now left with the normal users. We will need to first check to see if there are any groups that have been
    passed through. If no groups, then we check the max permission for those attributes.
    
    If the groups have been passed through, we will check to make sure that the end user has permissions for those 
    particular groups against those particular permission sets.
    """
    for row in permission_field:
        """
        We want to break out of this for loop if the entry is "". This should only occur when the user is looking at their
        profile. The user does not need permission to look at their profile.
        """
        if row == "":
            break


        #Users have no groups
        if group_list == None:
            #There is no group. Select the max value :)
            user_groups_results = user_group.objects.filter(
                is_deleted=False,
                username=request.user,
                permission_set__is_deleted=False,
            ).aggregate(Max('permission_set__' + row))
            user_permission_level[row] = user_groups_results['permission_set__' + row + '__max']
        else:
            #There is a group, lets find all permissions connected with this group :) and return the max :)
            #Default is 0
            group_permission = 0
            for group_id in group_list:
                #Grab user's permission for that group
                try:
                    user_groups_results = user_group.objects.filter(
                        is_deleted=False,
                        username=request.user,
                        permission_set__is_deleted=False,
                        group_id=group_id['group_id'],
                    ).aggregate(Max('permission_set__' + row))
                except:
                    user_groups_results = user_group.objects.filter(
                        is_deleted=False,
                        username=request.user,
                        permission_set__is_deleted=False,
                        group_id=group_id['group_id_id'],
                    ).aggregate(Max('permission_set__' + row))

                #Get the max value for the permission
                if not user_groups_results['permission_set__' + row + '__max'] == None:
                    if group_permission < user_groups_results['permission_set__' + row + '__max']:
                        group_permission = user_groups_results['permission_set__' + row + '__max']

            user_permission_level[row] = group_permission


    """
    The following code is for the menu. We will need to find out if a user can actually ADD any items and do any
    administration.
    """
    permission_results = user_group.objects.filter(
        is_deleted=False,
        username=request.user,
        permission_set__is_deleted=False,
    ).aggregate(
        Max('permission_set__project'),
        Max('permission_set__task'),
        Max('permission_set__requirement'),
        Max('permission_set__request_for_change'),
        Max('permission_set__organisation'),
        Max('permission_set__customer'),
        Max('permission_set__administration_assign_user_to_group'),
        Max('permission_set__administration_create_group'),
        Max('permission_set__administration_create_permission_set'),
        Max('permission_set__administration_create_user'),
    )

    user_permission_level['new_item'] = max(
        permission_results['permission_set__project__max'],
        permission_results['permission_set__task__max'],
        permission_results['permission_set__requirement__max'],
        permission_results['permission_set__organisation__max'],
        permission_results['permission_set__customer__max'],
    )
    user_permission_level['administration'] = max(
        permission_results['permission_set__administration_assign_user_to_group__max'],
        permission_results['permission_set__administration_create_group__max'],
        permission_results['permission_set__administration_create_permission_set__max'],
        permission_results['permission_set__administration_create_user__max'],
    )
    #END TEMP CODE

    return user_permission_level


from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min
from NearBeach.forms import *
from NearBeach.user_permissions import return_user_permission_level
from NearBeach.views.requirement_views import get_requirement_items
from django.views.decorators.http import require_http_methods

import json


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def add_requirement_item_link(request, requirement_item_id):
    # ADD IN PERMISSION
    permission_results = return_user_permission_level(request, None, ['requirement', 'requirement_link','requirement_item'])
    # What about requirement items? Will need to fix this elegantly.

    if permission_results['requirement'] < 2 & permission_results['requirement_item'] < 2:
        return HttpResponseRedirect(reverse('permission_denied'))

    form = AddRequirementLinkForm(request.POST)

    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the requirement instnace
    requirement_item_instance = requirement_item.objects.get(requirement_item_id=requirement_item_id)

    # Get the project list from the form
    for row in request.POST.getlist("project"):
        submit_object_assignment = object_assignment(
            requirement_item=requirement_item_instance,
            project=project.objects.get(project_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    for row in request.POST.getlist("task"):
        submit_object_assignment = object_assignment(
            requirement_item=requirement_item_instance,
            task=task.objects.get(task_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    for row in request.POST.getlist("opportunity"):
        submit_object_assignment = object_assignment(
            requirement_item=requirement_item_instance,
            opportunity=opportunity.objects.get(opportunity_id=row),
            change_user=request.user,
        )
        submit_object_assignment.save()

    return HttpResponse("Success")


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def get_requirement_item_links_list(request,requirement_item_id):
    """

    :param request:
    :param requirement_item_id:
    :return:
    """

    # Use object_assignment to get the requirme
    link_results = object_assignment.objects.filter(
        Q(
            is_deleted=False,
            requirement_item_id=requirement_item_id
        ) & Q(
            Q(opportunity_id__isnull=False) |
            Q(quote_id__isnull=False) |
            Q(project_id__isnull=False) |
            Q(task_id__isnull=False)
        )
    ).values(
        'opportunity_id',
        'opportunity_id__opportunity_name',
        'opportunity_id__opportunity_stage_id__opportunity_stage_description',
        'quote_id',
        'quote_id__quote_title',
        'quote_id__quote_stage_id__quote_stage',
        'project_id',
        'project_id__project_name',
        'project_id__project_status',
        'task_id',
        'task_id__task_short_description',
        'task_id__task_status',
        'requirement_item_id',
        'requirement_item_id__requirement_item_title',
    )

    """
    As explained on stack overflow here - https://stackoverflow.com/questions/7650448/django-serialize-queryset-values-into-json#31994176
    We need to Django's serializers can't handle a ValuesQuerySet. However, you can serialize by using a standard 
    json.dumps() and transforming your ValuesQuerySet to a list by using list().[sic]
    """

    # Send back json data
    json_results = json.dumps(list(link_results), cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login', redirect_field_name="")
def get_user_requirement_item_permissions(request, requirement_id):
    """
    Use the requirement_id and find out if the user has access to this requirement
    :param requirement_id:
    :return:
    """
    requirement_groups = object_assignment.objects.filter(
        is_deleted=False,
        # requirement_id=requirement_id
    ).values('group_id')

    if requirement_id > 0:
        # Make sure to filter by requirement groups
        requirement_groups = requirement_groups.filter(
            requirement_id=requirement_id
        )

    return return_user_permission_level(request, requirement_groups, ['requirement', 'requirement_link'])


@login_required(login_url='login', redirect_field_name="")
def new_requirement_item(request, requirement_id):
    # Check to see if POST
    if not request.method == "POST":
        return HttpResponseBadRequest("Sorry - needs to be in POST")

    # Check user permissions
    permission_results = get_user_requirement_item_permissions(request, requirement_id)

    # If user has no permissions to create requirements or requirement items, then send them to the appropriate location
    if permission_results['requirement'] <= 2 and permission_results['requirement_item'] <= 2:
        # Users can not create requirement.
        return HttpResponseRedirect(reverse('permission_denied'))

    # Get the data into the form for cleaning
    form = NewRequirementItemForm(request.POST)

    # Check to make sure there are no errors in the form
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    submit_requirement_item = requirement_item(
        requirement=requirement.objects.get(requirement_id=requirement_id),
        requirement_item_title=form.cleaned_data['requirement_item_title'],
        requirement_item_scope=form.cleaned_data['requirement_item_scope'],
        requirement_item_status=form.cleaned_data['requirement_item_status'],
        requirement_item_type=form.cleaned_data['requirement_item_type'],
        change_user=request.user,
    )
    submit_requirement_item.save()

    # Actuall return all the new requirement_item results to feed upstream
    return get_requirement_items(request, requirement_id)


@login_required(login_url='login', redirect_field_name="")
def requirement_item_information(request, requirement_item_id):
    """
        Loads the requirement item information.
        :param request:
        :param requirement_item_id:
        :return:
        """
    # Get the requirement information
    requirement_item_results = requirement_item.objects.get(requirement_item_id=requirement_item_id)

    # Check the permissions
    permission_results = get_user_requirement_item_permissions(request, requirement_item_id)

    # If user has no permissions to this requirement send them to the appropriate location
    if permission_results['requirement'] == 0:
        # Users who create the requirement get at least read only
        if requirement_item_results.creation_user == request.user:
            return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_item_id}))

        # Users who did not create the requirement get sent to permission denied.
        return HttpResponseRedirect(reverse('permission_denied'))

    # If the requirement has been closed - send user to the read only section
    if requirement_item_results.requirement_item_status.status_is_closed:
        return HttpResponseRedirect(reverse('requirement_readonly', args={requirement_item_id}))

    # Load template
    t = loader.get_template('NearBeach/requirement_items/requirement_item_information.html')

    # Get any extra data required
    organisation_results = organisation.objects.get(
        organisation_id=requirement_item_results.requirement.organisation_id,
    )

    status_list = list_of_requirement_item_status.objects.filter(
        is_deleted=False,
        status_is_closed=False,
    )

    type_list = list_of_requirement_item_type.objects.filter(
        is_deleted=False,
    )

    group_results = group.objects.filter(
        is_deleted=False,
    )

    # context
    c = {
        'group_results': serializers.serialize("json", group_results),
        'organisation_results': serializers.serialize("json", [organisation_results]),
        'permission_results': permission_results,
        'requirement_item_id': requirement_item_id,
        'requirement_item_results': serializers.serialize("json", [requirement_item_results]),
        'status_list': serializers.serialize("json", status_list),
        'type_list': serializers.serialize("json", type_list),
    }

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def requirement_information_save(request, requirement_item_id):
    """
    The following will save data
    :param request:
    :param requirement_id:
    :return:
    """
    # Check the permissions
    permission_results = get_user_requirement_item_permissions(request, requirement_item_id)

    # If user has no permissions to this requirement send them to the appropriate location
    if permission_results['requirement'] <= 1:
        return HttpResponseRedirect(reverse('permission_denied'))

    # Get form data
    form = UpdateRequirementItemForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    requirement_item_submit = requirement_item.objects.get(requirement_item_id=requirement_item_id)
    requirement_item_submit.change_user = request.user
    requirement_item_submit.requirement_item_title = form.cleaned_data['requirement_item_title']
    requirement_item_submit.requirement_item_scope = form.cleaned_data['requirement_item_scope']
    requirement_item_submit.requirement_item_status = form.cleaned_data['requirement_item_status']
    requirement_item_submit.requirement_item_type = form.cleaned_data['requirement_item_type']
    requirement_item_submit.save()

    # Send back an empty response
    return HttpResponse("")
